package com.sun.media.sound;

import javax.sound.midi.MetaMessage;
import javax.sound.midi.MidiDevice;
import javax.sound.midi.MidiEvent;
import java.util.WeakHashMap;
import javax.sound.midi.Transmitter;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.Synthesizer;
import javax.sound.midi.ControllerEventListener;
import javax.sound.midi.MetaEventListener;
import javax.sound.midi.Track;
import java.io.IOException;
import javax.sound.midi.MidiSystem;
import java.io.InputStream;
import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Receiver;
import java.util.ArrayList;
import java.util.List;
import javax.sound.midi.Sequence;
import java.util.Map;
import javax.sound.midi.Sequencer;

final class RealTimeSequencer extends AbstractMidiDevice implements Sequencer, AutoConnectSequencer
{
    private static final boolean DEBUG_PUMP = false;
    private static final boolean DEBUG_PUMP_ALL = false;
    private static final Map<ThreadGroup, EventDispatcher> dispatchers;
    static final RealTimeSequencer.RealTimeSequencerInfo info;
    private static final Sequencer.SyncMode[] masterSyncModes;
    private static final Sequencer.SyncMode[] slaveSyncModes;
    private static final Sequencer.SyncMode masterSyncMode;
    private static final Sequencer.SyncMode slaveSyncMode;
    private Sequence sequence;
    private double cacheTempoMPQ;
    private float cacheTempoFactor;
    private boolean[] trackMuted;
    private boolean[] trackSolo;
    private final MidiUtils.TempoCache tempoCache;
    private volatile boolean running;
    private RealTimeSequencer.PlayThread playThread;
    private volatile boolean recording;
    private final List recordingTracks;
    private long loopStart;
    private long loopEnd;
    private int loopCount;
    private final ArrayList metaEventListeners;
    private final ArrayList controllerEventListeners;
    private boolean autoConnect;
    private boolean doAutoConnectAtNextOpen;
    Receiver autoConnectedReceiver;
    
    RealTimeSequencer() throws MidiUnavailableException {
        super(RealTimeSequencer.info);
        this.sequence = null;
        this.cacheTempoMPQ = -1.0;
        this.cacheTempoFactor = -1.0f;
        this.trackMuted = null;
        this.trackSolo = null;
        this.tempoCache = new MidiUtils.TempoCache();
        this.recordingTracks = new ArrayList();
        this.loopStart = 0L;
        this.loopEnd = -1L;
        this.loopCount = 0;
        this.metaEventListeners = new ArrayList();
        this.controllerEventListeners = new ArrayList();
        this.autoConnect = false;
        this.doAutoConnectAtNextOpen = false;
        this.autoConnectedReceiver = null;
    }
    
    @Override
    public synchronized void setSequence(Sequence sequence) throws InvalidMidiDataException {
        if (sequence != this.sequence) {
            if (this.sequence != null && sequence == null) {
                this.setCaches();
                this.stop();
                this.trackMuted = null;
                this.trackSolo = null;
                this.loopStart = 0L;
                this.loopEnd = -1L;
                this.loopCount = 0;
                if (this.getDataPump() != null) {
                    this.getDataPump().setTickPos(0L);
                    this.getDataPump().resetLoopCount();
                }
            }
            if (this.playThread != null) {
                this.playThread.setSequence(sequence);
            }
            if ((this.sequence = sequence) != null) {
                this.tempoCache.refresh(sequence);
                this.setTickPosition(0L);
                this.propagateCaches();
            }
        }
        else if (sequence != null) {
            this.tempoCache.refresh(sequence);
            if (this.playThread != null) {
                this.playThread.setSequence(sequence);
            }
        }
    }
    
    @Override
    public synchronized void setSequence(InputStream stream) throws IOException, InvalidMidiDataException {
        if (stream == null) {
            this.setSequence((Sequence)null);
            return;
        }
        this.setSequence(MidiSystem.getSequence(stream));
        return;
    }
    
    @Override
    public Sequence getSequence() {
        return this.sequence;
    }
    
    @Override
    public synchronized void start() {
        if (!this.isOpen()) {
            throw new IllegalStateException("sequencer not open");
        }
        if (this.sequence == null) {
            throw new IllegalStateException("sequence not set");
        }
        if (this.running) {
            return;
        }
        this.implStart();
        return;
    }
    
    @Override
    public synchronized void stop() {
        if (!this.isOpen()) {
            throw new IllegalStateException("sequencer not open");
        }
        this.stopRecording();
        if (!this.running) {
            return;
        }
        this.implStop();
        return;
    }
    
    @Override
    public boolean isRunning() {
        return this.running;
    }
    
    @Override
    public void startRecording() {
        if (!this.isOpen()) {
            throw new IllegalStateException("Sequencer not open");
        }
        this.start();
        this.recording = true;
        return;
    }
    
    @Override
    public void stopRecording() {
        if (!this.isOpen()) {
            throw new IllegalStateException("Sequencer not open");
        }
        this.recording = false;
        return;
    }
    
    @Override
    public boolean isRecording() {
        return this.recording;
    }
    
    @Override
    public void recordEnable(Track track, int n) {
        if (!this.findTrack(track)) {
            throw new IllegalArgumentException("Track does not exist in the current sequence");
        }
        synchronized (this.recordingTracks) {
            RealTimeSequencer.RecordingTrack value = RealTimeSequencer.RecordingTrack.get(this.recordingTracks, track);
            if (value != null) {
                value.channel = n;
            }
            else {
                this.recordingTracks.add(new RealTimeSequencer.RecordingTrack(track, n));
            }
        }
        return;
    }
    
    @Override
    public void recordDisable(Track track) {
        synchronized (this.recordingTracks) {
            RealTimeSequencer.RecordingTrack value = RealTimeSequencer.RecordingTrack.get(this.recordingTracks, track);
            if (value != null) {
                this.recordingTracks.remove(value);
            }
        }
    }
    
    private boolean findTrack(Track track) {
        boolean n = false;
        if (this.sequence != null) {
            Track[] tracks = this.sequence.getTracks();
            int i = 0;
            while (i < tracks.length) {
                if (track == tracks[i]) {
                    n = true;
                    break;
                }
                ++i;
            }
        }
        return n;
    }
    
    @Override
    public float getTempoInBPM() {
        return (float)MidiUtils.convertTempo((double)this.getTempoInMPQ());
    }
    
    @Override
    public void setTempoInBPM(float n) {
        if (n <= 0.0f) {
            n = 1.0f;
        }
        this.setTempoInMPQ((float)MidiUtils.convertTempo((double)n));
    }
    
    @Override
    public float getTempoInMPQ() {
        if (this.needCaching()) {
            if (this.cacheTempoMPQ != -1.0) {
                return (float)this.cacheTempoMPQ;
            }
            if (this.sequence != null) {
                return this.tempoCache.getTempoMPQAt(this.getTickPosition());
            }
            return 500000.0f;
        }
        else {
            return this.getDataPump().getTempoMPQ();
        }
    }
    
    @Override
    public void setTempoInMPQ(float tempoMPQ) {
        if (tempoMPQ <= 0.0f) {
            tempoMPQ = 1.0f;
        }
        if (this.needCaching()) {
            this.cacheTempoMPQ = (double)tempoMPQ;
        }
        else {
            this.getDataPump().setTempoMPQ(tempoMPQ);
            this.cacheTempoMPQ = -1.0;
        }
    }
    
    @Override
    public void setTempoFactor(float n) {
        if (n <= 0.0f) {
            return;
        }
        if (this.needCaching()) {
            this.cacheTempoFactor = n;
        }
        else {
            this.getDataPump().setTempoFactor(n);
            this.cacheTempoFactor = -1.0f;
        }
        return;
    }
    
    @Override
    public float getTempoFactor() {
        if (this.needCaching()) {
            if (this.cacheTempoFactor != -1.0f) {
                return this.cacheTempoFactor;
            }
            return 1.0f;
        }
        else {
            return this.getDataPump().getTempoFactor();
        }
    }
    
    @Override
    public long getTickLength() {
        if (this.sequence == null) {
            return 0L;
        }
        return this.sequence.getTickLength();
    }
    
    @Override
    public synchronized long getTickPosition() {
        if (this.getDataPump() == null || this.sequence == null) {
            return 0L;
        }
        return this.getDataPump().getTickPos();
    }
    
    @Override
    public synchronized void setTickPosition(long tickPos) {
        if (tickPos < 0L) {
            return;
        }
        if (this.getDataPump() == null) {
            if (tickPos == 0L) {}
        }
        else if (this.sequence == null) {
            if (tickPos == 0L) {}
        }
        else {
            this.getDataPump().setTickPos(tickPos);
        }
        return;
    }
    
    @Override
    public long getMicrosecondLength() {
        if (this.sequence == null) {
            return 0L;
        }
        return this.sequence.getMicrosecondLength();
    }
    
    @Override
    public long getMicrosecondPosition() {
        if (this.getDataPump() == null || this.sequence == null) {
            return 0L;
        }
        synchronized (this.tempoCache) {
            return MidiUtils.tick2microsecond(this.sequence, this.getDataPump().getTickPos(), this.tempoCache);
        }
    }
    
    @Override
    public void setMicrosecondPosition(long n) {
        if (n < 0L) {
            return;
        }
        if (this.getDataPump() == null) {
            if (n == 0L) {}
        }
        else if (this.sequence == null) {
            if (n == 0L) {}
        }
        else {
            synchronized (this.tempoCache) {
                this.setTickPosition(MidiUtils.microsecond2tick(this.sequence, n, this.tempoCache));
            }
        }
        return;
    }
    
    @Override
    public void setMasterSyncMode(Sequencer.SyncMode syncMode) {
    }
    
    @Override
    public Sequencer.SyncMode getMasterSyncMode() {
        return RealTimeSequencer.masterSyncMode;
    }
    
    @Override
    public Sequencer.SyncMode[] getMasterSyncModes() {
        Sequencer.SyncMode[] array = new Sequencer.SyncMode[RealTimeSequencer.masterSyncModes.length];
        System.arraycopy(RealTimeSequencer.masterSyncModes, 0, array, 0, RealTimeSequencer.masterSyncModes.length);
        return array;
    }
    
    @Override
    public void setSlaveSyncMode(Sequencer.SyncMode syncMode) {
    }
    
    @Override
    public Sequencer.SyncMode getSlaveSyncMode() {
        return RealTimeSequencer.slaveSyncMode;
    }
    
    @Override
    public Sequencer.SyncMode[] getSlaveSyncModes() {
        Sequencer.SyncMode[] array = new Sequencer.SyncMode[RealTimeSequencer.slaveSyncModes.length];
        System.arraycopy(RealTimeSequencer.slaveSyncModes, 0, array, 0, RealTimeSequencer.slaveSyncModes.length);
        return array;
    }
    
    int getTrackCount() {
        if (this.getSequence() != null) {
            return this.sequence.getTracks().length;
        }
        return 0;
    }
    
    @Override
    public synchronized void setTrackMute(int n, boolean b) {
        int trackCount = this.getTrackCount();
        if (n < 0 || n >= this.getTrackCount()) {
            return;
        }
        (this.trackMuted = ensureBoolArraySize(this.trackMuted, trackCount))[n] = b;
        if (this.getDataPump() != null) {
            this.getDataPump().muteSoloChanged();
        }
        return;
    }
    
    @Override
    public synchronized boolean getTrackMute(int n) {
        return n >= 0 && n < this.getTrackCount() && this.trackMuted != null && this.trackMuted.length > n && this.trackMuted[n];
    }
    
    @Override
    public synchronized void setTrackSolo(int n, boolean b) {
        int trackCount = this.getTrackCount();
        if (n < 0 || n >= this.getTrackCount()) {
            return;
        }
        (this.trackSolo = ensureBoolArraySize(this.trackSolo, trackCount))[n] = b;
        if (this.getDataPump() != null) {
            this.getDataPump().muteSoloChanged();
        }
        return;
    }
    
    @Override
    public synchronized boolean getTrackSolo(int n) {
        return n >= 0 && n < this.getTrackCount() && this.trackSolo != null && this.trackSolo.length > n && this.trackSolo[n];
    }
    
    @Override
    public boolean addMetaEventListener(MetaEventListener metaEventListener) {
        synchronized (this.metaEventListeners) {
            if (!this.metaEventListeners.contains(metaEventListener)) {
                this.metaEventListeners.add(metaEventListener);
            }
            return true;
        }
    }
    
    @Override
    public void removeMetaEventListener(MetaEventListener o) {
        synchronized (this.metaEventListeners) {
            int index = this.metaEventListeners.indexOf(o);
            if (index >= 0) {
                this.metaEventListeners.remove(index);
            }
        }
    }
    
    @Override
    public int[] addControllerEventListener(ControllerEventListener obj, int[] array) {
        synchronized (this.controllerEventListeners) {
            Object e = null;
            boolean n = false;
            int i = 0;
            while (i < this.controllerEventListeners.size()) {
                e = (Object)this.controllerEventListeners.get(i);
                if (e.listener.equals(obj)) {
                    RealTimeSequencer.ControllerListElement.access$200(e, array);
                    n = true;
                    break;
                }
                ++i;
            }
            if (n == 0) {
                e = new RealTimeSequencer.ControllerListElement(obj, array);
                this.controllerEventListeners.add(e);
            }
            return RealTimeSequencer.ControllerListElement.access$400(e);
        }
    }
    
    @Override
    public int[] removeControllerEventListener(ControllerEventListener obj, int[] array) {
        synchronized (this.controllerEventListeners) {
            RealTimeSequencer.ControllerListElement o = null;
            boolean n = false;
            int i = 0;
            while (i < this.controllerEventListeners.size()) {
                o = (RealTimeSequencer.ControllerListElement)this.controllerEventListeners.get(i);
                if (o.listener.equals(obj)) {
                    RealTimeSequencer.ControllerListElement.access$500(o, array);
                    n = true;
                    break;
                }
                ++i;
            }
            if (n == 0) {
                return new int[0];
            }
            if (array == null) {
                int index = this.controllerEventListeners.indexOf(o);
                if (index >= 0) {
                    this.controllerEventListeners.remove(index);
                }
                return new int[0];
            }
            else {
                return RealTimeSequencer.ControllerListElement.access$400(o);
            }
        }
    }
    
    @Override
    public void setLoopStartPoint(long n) {
        if (n > this.getTickLength() || (this.loopEnd != -1L && n > this.loopEnd) || n < 0L) {
            throw new IllegalArgumentException("invalid loop start point: " + n);
        }
        this.loopStart = n;
        return;
    }
    
    @Override
    public long getLoopStartPoint() {
        return this.loopStart;
    }
    
    @Override
    public void setLoopEndPoint(long n) {
        if (n > this.getTickLength() || (this.loopStart > n && n != -1L) || n < -1L) {
            throw new IllegalArgumentException("invalid loop end point: " + n);
        }
        this.loopEnd = n;
        return;
    }
    
    @Override
    public long getLoopEndPoint() {
        return this.loopEnd;
    }
    
    @Override
    public void setLoopCount(int n) {
        if (n != -1 && n < 0) {
            throw new IllegalArgumentException("illegal value for loop count: " + n);
        }
        this.loopCount = n;
        if (this.getDataPump() != null) {
            this.getDataPump().resetLoopCount();
        }
        return;
    }
    
    @Override
    public int getLoopCount() {
        return this.loopCount;
    }
    
    @Override
    protected void implOpen() throws MidiUnavailableException {
        this.playThread = new RealTimeSequencer.PlayThread();
        if (this.sequence != null) {
            this.playThread.setSequence(this.sequence);
        }
        this.propagateCaches();
        if (this.doAutoConnectAtNextOpen) {
            this.doAutoConnect();
        }
    }
    
    private void doAutoConnect() {
        Receiver receiver = null;
        try {
            Synthesizer synthesizer = MidiSystem.getSynthesizer();
            if ((synthesizer instanceof ReferenceCountingDevice)) {
                receiver = ((ReferenceCountingDevice)synthesizer).getReceiverReferenceCounting();
            }
            else {
                synthesizer.open();
                try {
                    receiver = synthesizer.getReceiver();
                }
                finally {
                    if (receiver == null) {
                        synthesizer.close();
                    }
                }
            }
        }
        catch (Exception ex) {}
        if (receiver == null) {
            try {
                receiver = MidiSystem.getReceiver();
            }
            catch (Exception ex2) {}
        }
        if (receiver != null) {
            this.autoConnectedReceiver = receiver;
            try {
                this.getTransmitter().setReceiver(receiver);
            }
            catch (Exception ex3) {}
        }
    }
    
    private synchronized void propagateCaches() {
        if (this.sequence != null && this.isOpen()) {
            if (this.cacheTempoFactor != -1.0f) {
                this.setTempoFactor(this.cacheTempoFactor);
            }
            if (this.cacheTempoMPQ == -1.0) {
                this.setTempoInMPQ(new MidiUtils.TempoCache(this.sequence).getTempoMPQAt(this.getTickPosition()));
            }
            else {
                this.setTempoInMPQ((float)this.cacheTempoMPQ);
            }
        }
    }
    
    private synchronized void setCaches() {
        this.cacheTempoFactor = this.getTempoFactor();
        this.cacheTempoMPQ = (double)this.getTempoInMPQ();
    }
    
    @Override
    protected synchronized void implClose() {
        if (this.playThread != null) {
            this.playThread.close();
            this.playThread = null;
        }
        super.implClose();
        this.sequence = null;
        this.running = false;
        this.cacheTempoMPQ = -1.0;
        this.cacheTempoFactor = -1.0f;
        this.trackMuted = null;
        this.trackSolo = null;
        this.loopStart = 0L;
        this.loopEnd = -1L;
        this.loopCount = 0;
        this.doAutoConnectAtNextOpen = this.autoConnect;
        if (this.autoConnectedReceiver != null) {
            try {
                this.autoConnectedReceiver.close();
            }
            catch (Exception ex) {}
            this.autoConnectedReceiver = null;
        }
    }
    
    void implStart() {
        if (this.playThread == null) {
            return;
        }
        this.tempoCache.refresh(this.sequence);
        if (!this.running) {
            this.running = true;
            this.playThread.start();
        }
        return;
    }
    
    void implStop() {
        if (this.playThread == null) {
            return;
        }
        this.recording = false;
        if (this.running) {
            this.running = false;
            this.playThread.stop();
        }
        return;
    }
    
    private static EventDispatcher getEventDispatcher() {
        ThreadGroup threadGroup = Thread.currentThread().getThreadGroup();
        synchronized (RealTimeSequencer.dispatchers) {
            EventDispatcher eventDispatcher = (EventDispatcher)RealTimeSequencer.dispatchers.get(threadGroup);
            if (eventDispatcher == null) {
                eventDispatcher = new EventDispatcher();
                RealTimeSequencer.dispatchers.put(threadGroup, eventDispatcher);
                eventDispatcher.start();
            }
            return eventDispatcher;
        }
    }
    
    void sendMetaEvents(MidiMessage midiMessage) {
        if (this.metaEventListeners.size() == 0) {
            return;
        }
        getEventDispatcher().sendAudioEvents(midiMessage, this.metaEventListeners);
        return;
    }
    
    void sendControllerEvents(MidiMessage midiMessage) {
        int size = this.controllerEventListeners.size();
        if (size == 0) {
            return;
        }
        if (!(midiMessage instanceof ShortMessage)) {
            return;
        }
        int data1 = ((ShortMessage)midiMessage).getData1();
        ArrayList<RealTimeSequencer.ControllerListElement> list = new ArrayList();
        for (int i = 0; i < size; ++i) {
            RealTimeSequencer.ControllerListElement controllerListElement = (RealTimeSequencer.ControllerListElement)this.controllerEventListeners.get(i);
            int j = 0;
            while (j < controllerListElement.controllers.length) {
                if (controllerListElement.controllers[j] == data1) {
                    list.add(controllerListElement.listener);
                    break;
                }
                ++j;
            }
        }
        getEventDispatcher().sendAudioEvents(midiMessage, list);
        return;
    }
    
    private boolean needCaching() {
        return !this.isOpen() || this.sequence == null || this.playThread == null;
    }
    
    private RealTimeSequencer.DataPump getDataPump() {
        if (this.playThread != null) {
            return this.playThread.getDataPump();
        }
        return null;
    }
    
    private MidiUtils.TempoCache getTempoCache() {
        return this.tempoCache;
    }
    
    private static boolean[] ensureBoolArraySize(boolean[] array, int n) {
        if (array == null) {
            return new boolean[n];
        }
        if (array.length < n) {
            boolean[] array2 = new boolean[n];
            System.arraycopy(array, 0, array2, 0, array.length);
            return array2;
        }
        return array;
    }
    
    @Override
    protected boolean hasReceivers() {
        return true;
    }
    
    @Override
    protected Receiver createReceiver() throws MidiUnavailableException {
        return new RealTimeSequencer.SequencerReceiver();
    }
    
    @Override
    protected boolean hasTransmitters() {
        return true;
    }
    
    @Override
    protected Transmitter createTransmitter() throws MidiUnavailableException {
        return new RealTimeSequencer.SequencerTransmitter();
    }
    
    @Override
    public void setAutoConnect(Receiver autoConnectedReceiver) {
        this.autoConnect = (autoConnectedReceiver != null);
        this.autoConnectedReceiver = autoConnectedReceiver;
    }
    
    static MidiUtils.TempoCache access$1400(RealTimeSequencer realTimeSequencer) {
        return realTimeSequencer.getTempoCache();
    }
    
    static {
        dispatchers = new WeakHashMap();
        info = new RealTimeSequencer.RealTimeSequencerInfo();
        masterSyncModes = new Sequencer.SyncMode[] { Sequencer.SyncMode.INTERNAL_CLOCK };
        slaveSyncModes = new Sequencer.SyncMode[] { Sequencer.SyncMode.NO_SYNC };
        masterSyncMode = Sequencer.SyncMode.INTERNAL_CLOCK;
        slaveSyncMode = Sequencer.SyncMode.NO_SYNC;
    }
    
    private class SequencerTransmitter extends AbstractMidiDevice.BasicTransmitter
    {
    }
    
    final class SequencerReceiver extends AbstractMidiDevice.AbstractReceiver
    {
        @Override
        void implSend(MidiMessage message, long n) {
            if (RealTimeSequencer.this.recording) {
                long tick = 0L;
                if (n < 0L) {
                    tick = RealTimeSequencer.this.getTickPosition();
                }
                else {
                    synchronized (RealTimeSequencer.this.tempoCache) {
                        tick = MidiUtils.microsecond2tick(RealTimeSequencer.this.sequence, n, RealTimeSequencer.this.tempoCache);
                    }
                }
                Track track = null;
                if (message.getLength() > 1) {
                    if ((message instanceof ShortMessage)) {
                        ShortMessage shortMessage = (ShortMessage)message;
                        if ((shortMessage.getStatus() & 0xF0) != 0xF0) {
                            track = RealTimeSequencer.RecordingTrack.get(RealTimeSequencer.this.recordingTracks, shortMessage.getChannel());
                        }
                    }
                    else {
                        track = RealTimeSequencer.RecordingTrack.get(RealTimeSequencer.this.recordingTracks, -1);
                    }
                    if (track != null) {
                        if ((message instanceof ShortMessage)) {
                            message = new FastShortMessage((ShortMessage)message);
                        }
                        else {
                            message = (Object)message.clone();
                        }
                        track.add(new MidiEvent(message, tick));
                    }
                }
            }
        }
    }
    
    private static class RealTimeSequencerInfo extends MidiDevice.Info
    {
        private static final String name = "Real Time Sequencer";
        private static final String vendor = "Oracle Corporation";
        private static final String description = "Software sequencer";
        private static final String version = "Version 1.0";
        
        private RealTimeSequencerInfo() {
            super("Real Time Sequencer", "Oracle Corporation", "Software sequencer", "Version 1.0");
        }
    }
    
    private class ControllerListElement
    {
        int[] controllers;
        final ControllerEventListener listener;
        
        private ControllerListElement(ControllerEventListener listener, int[] controllers) {
            this.listener = listener;
            if (controllers == null) {
                controllers = new int[128];
                for (int i = 0; i < 128; ++i) {
                    controllers[i] = i;
                }
            }
            this.controllers = controllers;
        }
        
        private void addControllers(int[] array) {
            if (array == null) {
                this.controllers = new int[128];
                for (int n2 = 0; n2 < 128; ++n2) {
                    this.controllers[n2] = n2;
                }
                return;
            }
            int[] array2 = new int[this.controllers.length + array.length];
            for (int i = 0; i < this.controllers.length; ++i) {
                array2[i] = this.controllers[i];
            }
            int length = this.controllers.length;
            for (int j = 0; j < array.length; ++j) {
                boolean n = false;
                int k = 0;
                while (k < this.controllers.length) {
                    if (array[j] == this.controllers[k]) {
                        n = true;
                        break;
                    }
                    ++k;
                }
                if (n == 0) {
                    array2[length++] = array[j];
                }
            }
            int[] controllers = new int[length];
            for (int l = 0; l < length; ++l) {
                controllers[l] = array2[l];
            }
            this.controllers = controllers;
            return;
        }
        
        private void removeControllers(int[] array) {
            if (array == null) {
                this.controllers = new int[0];
            }
            else {
                int[] array2 = new int[this.controllers.length];
                int n = 0;
                for (int i = 0; i < this.controllers.length; ++i) {
                    boolean n2 = false;
                    int j = 0;
                    while (j < array.length) {
                        if (this.controllers[i] == array[j]) {
                            n2 = true;
                            break;
                        }
                        ++j;
                    }
                    if (n2 == 0) {
                        array2[n++] = this.controllers[i];
                    }
                }
                int[] controllers = new int[n];
                for (int k = 0; k < n; ++k) {
                    controllers[k] = array2[k];
                }
                this.controllers = controllers;
            }
        }
        
        private int[] getControllers() {
            if (this.controllers == null) {
                return null;
            }
            int[] array = new int[this.controllers.length];
            for (int i = 0; i < this.controllers.length; ++i) {
                array[i] = this.controllers[i];
            }
            return array;
        }
        
        static void access$200(RealTimeSequencer.ControllerListElement controllerListElement, int[] array) {
            controllerListElement.addControllers(array);
        }
        
        static int[] access$400(RealTimeSequencer.ControllerListElement controllerListElement) {
            return controllerListElement.getControllers();
        }
        
        static void access$500(RealTimeSequencer.ControllerListElement controllerListElement, int[] array) {
            controllerListElement.removeControllers(array);
        }
    }
    
    static class RecordingTrack
    {
        private final Track track;
        private int channel;
        
        RecordingTrack(Track track, int channel) {
            this.track = track;
            this.channel = channel;
        }
        
        static RealTimeSequencer.RecordingTrack get(List list, Track track) {
            synchronized (list) {
                int size = list.size();
                int i = 0;
                while (i < size) {
                    RealTimeSequencer.RecordingTrack recordingTrack = (RealTimeSequencer.RecordingTrack)list.get(i);
                    if (recordingTrack.track == track) {
                        return recordingTrack;
                    }
                    ++i;
                }
            }
            return null;
        }
        
        static Track get(List list, int n) {
            synchronized (list) {
                int size = list.size();
                int i = 0;
                while (i < size) {
                    RealTimeSequencer.RecordingTrack recordingTrack = (RealTimeSequencer.RecordingTrack)list.get(i);
                    if (recordingTrack.channel == n || recordingTrack.channel == -1) {
                        return recordingTrack.track;
                    }
                    ++i;
                }
            }
            return null;
        }
    }
    
    final class PlayThread implements Runnable
    {
        private Thread thread;
        private final Object lock;
        boolean interrupted;
        boolean isPumping;
        private final RealTimeSequencer.DataPump dataPump;
        
        PlayThread() {
            this.lock = new Object();
            this.interrupted = false;
            this.isPumping = false;
            this.dataPump = new RealTimeSequencer.DataPump();
            this.thread = JSSecurityManager.createThread(this, "Java Sound Sequencer", false, 8, true);
        }
        
        RealTimeSequencer.DataPump getDataPump() {
            return this.dataPump;
        }
        
        synchronized void setSequence(Sequence sequence) {
            this.dataPump.setSequence(sequence);
        }
        
        synchronized void start() {
            RealTimeSequencer.this.running = true;
            if (!this.dataPump.hasCachedTempo()) {
                this.dataPump.setTempoMPQ(RealTimeSequencer.this.tempoCache.getTempoMPQAt(RealTimeSequencer.this.getTickPosition()));
            }
            this.dataPump.checkPointMillis = 0L;
            this.dataPump.clearNoteOnCache();
            this.dataPump.needReindex = true;
            this.dataPump.resetLoopCount();
            synchronized (this.lock) {
                this.lock.notifyAll();
            }
        }
        
        synchronized void stop() {
            this.playThreadImplStop();
            long n = System.nanoTime() / 1000000L;
            while (this.isPumping) {
                synchronized (this.lock) {
                    try {
                        this.lock.wait(2000L);
                    }
                    catch (InterruptedException ex) {}
                }
                if (System.nanoTime() / 1000000L - n > 1900L) {
                    continue;
                }
                continue;
            }
        }
        
        void playThreadImplStop() {
            RealTimeSequencer.this.running = false;
            synchronized (this.lock) {
                this.lock.notifyAll();
            }
        }
        
        void close() {
            Thread thread = null;
            synchronized (this) {
                this.interrupted = true;
                thread = this.thread;
                this.thread = null;
            }
            if (thread != null) {
                synchronized (this.lock) {
                    this.lock.notifyAll();
                }
            }
            if (thread != null) {
                try {
                    thread.join(2000L);
                }
                catch (InterruptedException ex) {}
            }
        }
        
        @Override
        public void run() {
            while (!this.interrupted) {
                boolean pump = false;
                int access$1100 = RealTimeSequencer.this.running;
                this.isPumping = (!this.interrupted && RealTimeSequencer.this.running);
                while (pump == 0 && !this.interrupted && RealTimeSequencer.this.running) {
                    pump = this.dataPump.pump();
                    try {
                        Thread.sleep(1L);
                    }
                    catch (InterruptedException ex) {}
                }
                this.playThreadImplStop();
                if (!(access$1100 == 0)) {
                    this.dataPump.notesOff(true);
                }
                if (!(pump == 0)) {
                    this.dataPump.setTickPos(RealTimeSequencer.this.sequence.getTickLength());
                    MetaMessage o = new MetaMessage();
                    try {
                        o.setMessage(47, new byte[0], 0);
                    }
                    catch (InvalidMidiDataException ex2) {}
                    RealTimeSequencer.this.sendMetaEvents(o);
                }
                synchronized (this.lock) {
                    this.isPumping = false;
                    this.lock.notifyAll();
                    while (!RealTimeSequencer.this.running && !this.interrupted) {
                        try {
                            this.lock.wait();
                        }
                        catch (Exception ex3) {}
                    }
                }
            }
        }
    }
    
    private class DataPump
    {
        private float currTempo;
        private float tempoFactor;
        private float inverseTempoFactor;
        private long ignoreTempoEventAt;
        private int resolution;
        private float divisionType;
        private long checkPointMillis;
        private long checkPointTick;
        private int[] noteOnCache;
        private Track[] tracks;
        private boolean[] trackDisabled;
        private int[] trackReadPos;
        private long lastTick;
        private boolean needReindex;
        private int currLoopCounter;
        
        DataPump() {
            this.needReindex = false;
            this.currLoopCounter = 0;
            this.init();
        }
        
        synchronized void init() {
            this.ignoreTempoEventAt = -1L;
            this.tempoFactor = 1.0f;
            this.inverseTempoFactor = 1.0f;
            this.noteOnCache = new int[128];
            this.tracks = null;
            this.trackDisabled = null;
        }
        
        synchronized void setTickPos(long lastTick) {
            this.lastTick = lastTick;
            if (RealTimeSequencer.this.running) {
                this.notesOff(false);
            }
            if (RealTimeSequencer.this.running || lastTick > 0L) {
                this.chaseEvents(lastTick, lastTick);
            }
            else {
                this.needReindex = true;
            }
            if (!this.hasCachedTempo()) {
                this.setTempoMPQ(RealTimeSequencer.access$1400(RealTimeSequencer.this).getTempoMPQAt(this.lastTick, this.currTempo));
                this.ignoreTempoEventAt = -1L;
            }
            this.checkPointMillis = 0L;
        }
        
        long getTickPos() {
            return this.lastTick;
        }
        
        boolean hasCachedTempo() {
            if (this.ignoreTempoEventAt != this.lastTick) {
                this.ignoreTempoEventAt = -1L;
            }
            return this.ignoreTempoEventAt >= 0L;
        }
        
        synchronized void setTempoMPQ(float currTempo) {
            if (currTempo > 0.0f && currTempo != this.currTempo) {
                this.ignoreTempoEventAt = this.lastTick;
                this.currTempo = currTempo;
                this.checkPointMillis = 0L;
            }
        }
        
        float getTempoMPQ() {
            return this.currTempo;
        }
        
        synchronized void setTempoFactor(float tempoFactor) {
            if (tempoFactor > 0.0f && tempoFactor != this.tempoFactor) {
                this.tempoFactor = tempoFactor;
                this.inverseTempoFactor = 1.0f / tempoFactor;
                this.checkPointMillis = 0L;
            }
        }
        
        float getTempoFactor() {
            return this.tempoFactor;
        }
        
        synchronized void muteSoloChanged() {
            boolean[] disabledArray = this.makeDisabledArray();
            if (RealTimeSequencer.this.running) {
                this.applyDisabledTracks(this.trackDisabled, disabledArray);
            }
            this.trackDisabled = disabledArray;
        }
        
        synchronized void setSequence(Sequence sequence) {
            if (sequence == null) {
                this.init();
                return;
            }
            this.tracks = sequence.getTracks();
            this.muteSoloChanged();
            this.resolution = sequence.getResolution();
            this.divisionType = sequence.getDivisionType();
            this.trackReadPos = new int[this.tracks.length];
            this.checkPointMillis = 0L;
            this.needReindex = true;
            return;
        }
        
        synchronized void resetLoopCount() {
            this.currLoopCounter = RealTimeSequencer.this.loopCount;
        }
        
        void clearNoteOnCache() {
            for (int i = 0; i < 128; ++i) {
                this.noteOnCache[i] = 0;
            }
        }
        
        void notesOff(boolean b) {
            int n = 0;
            for (int i = 0; i < 16; ++i) {
                int n2 = 1 << i;
                for (int j = 0; j < 128; ++j) {
                    if ((this.noteOnCache[j] & n2) != 0x0) {
                        int[] noteOnCache = this.noteOnCache;
                        int n3 = j;
                        noteOnCache[n3] ^= n2;
                        RealTimeSequencer.this.getTransmitterList().sendMessage(0x90 | i | j << 8, -1L);
                        ++n;
                    }
                }
                RealTimeSequencer.this.getTransmitterList().sendMessage(0xB0 | i | 0x7B00, -1L);
                RealTimeSequencer.this.getTransmitterList().sendMessage(0xB0 | i | 0x4000, -1L);
                if (b) {
                    RealTimeSequencer.this.getTransmitterList().sendMessage(0xB0 | i | 0x7900, -1L);
                    ++n;
                }
            }
        }
        
        private boolean[] makeDisabledArray() {
            if (this.tracks == null) {
                return null;
            }
            boolean[] array = new boolean[this.tracks.length];
            boolean[] access$1600;
            boolean[] access$1601;
            synchronized (RealTimeSequencer.this) {
                access$1600 = RealTimeSequencer.this.trackMuted;
                access$1601 = RealTimeSequencer.this.trackSolo;
            }
            boolean n = false;
            if (access$1601 != null) {
                int i = 0;
                while (i < access$1601.length) {
                    if (access$1601[i]) {
                        n = true;
                        break;
                    }
                    ++i;
                }
            }
            if (!(n == 0)) {
                for (int k = 0; k < array.length; ++k) {
                    array[k] = (k >= access$1601.length || !access$1601[k]);
                }
                return array;
            }
            for (int j = 0; j < array.length; ++j) {
                array[j] = (access$1600 != null && j < access$1600.length && access$1600[j]);
            }
            return array;
        }
        
        private void sendNoteOffIfOn(Track track, long n) {
            int size = track.size();
            int n2 = 0;
            try {
                int i = 0;
                while (i < size) {
                    MidiEvent value = track.get(i);
                    if (value.getTick() > n) {
                        break;
                    }
                    MidiMessage message = value.getMessage();
                    int status = message.getStatus();
                    if (message.getLength() == 3 && (status & 0xF0) == 0x90) {
                        int data1 = -1;
                        if ((message instanceof ShortMessage)) {
                            ShortMessage shortMessage = (ShortMessage)message;
                            if (shortMessage.getData2() > 0) {
                                data1 = shortMessage.getData1();
                            }
                        }
                        else {
                            byte[] message2 = message.getMessage();
                            if ((message2[2] & 0x7F) > 0) {
                                data1 = (message2[1] & 0x7F);
                            }
                        }
                        if (data1 >= 0) {
                            int n3 = 1 << (status & 0xF);
                            if ((this.noteOnCache[data1] & n3) != 0x0) {
                                RealTimeSequencer.this.getTransmitterList().sendMessage(status | data1 << 8, -1L);
                                int[] noteOnCache = this.noteOnCache;
                                int n4 = data1;
                                noteOnCache[n4] &= (0xFFFF ^ n3);
                                ++n2;
                            }
                        }
                    }
                    ++i;
                }
            }
            catch (ArrayIndexOutOfBoundsException ex) {}
        }
        
        private void applyDisabledTracks(boolean[] array, boolean[] array2) {
            byte[][] array3 = (byte[][])null;
            synchronized (RealTimeSequencer.this) {
                for (int i = 0; i < array2.length; ++i) {
                    if ((array == null || i >= array.length || !array[i]) && array2[i]) {
                        if (this.tracks.length > i) {
                            this.sendNoteOffIfOn(this.tracks[i], this.lastTick);
                        }
                    }
                    else if (array != null && i < array.length && array[i] && !array2[i]) {
                        if (array3 == null) {
                            array3 = new byte[128][16];
                        }
                        this.chaseTrackEvents(i, 0L, this.lastTick, true, array3);
                    }
                }
            }
        }
        
        private void chaseTrackEvents(int n, long n2, long n3, boolean b, byte[][] array) {
            if (n2 > n3) {
                n2 = 0L;
            }
            byte[] array2 = new byte[16];
            for (int i = 0; i < 16; ++i) {
                array2[i] = -1;
                for (int j = 0; j < 128; ++j) {
                    array[j][i] = -1;
                }
            }
            Track track = this.tracks[n];
            int size = track.size();
            try {
                int k = 0;
                while (k < size) {
                    MidiEvent value = track.get(k);
                    if (value.getTick() >= n3) {
                        if (b && n < this.trackReadPos.length) {
                            this.trackReadPos[n] = ((k > 0) ? (k - 1) : 0);
                            break;
                        }
                        break;
                    }
                    else {
                        MidiMessage message = value.getMessage();
                        int status = message.getStatus();
                        int length = message.getLength();
                        if (length == 3 && (status & 0xF0) == 0xB0) {
                            if ((message instanceof ShortMessage)) {
                                ShortMessage shortMessage = (ShortMessage)message;
                                array[shortMessage.getData1() & 0x7F][status & 0xF] = (byte)shortMessage.getData2();
                            }
                            else {
                                byte[] message2 = message.getMessage();
                                array[message2[1] & 0x7F][status & 0xF] = message2[2];
                            }
                        }
                        if (length == 2 && (status & 0xF0) == 0xC0) {
                            if ((message instanceof ShortMessage)) {
                                array2[status & 0xF] = (byte)((ShortMessage)message).getData1();
                            }
                            else {
                                array2[status & 0xF] = message.getMessage()[1];
                            }
                        }
                        ++k;
                    }
                }
            }
            catch (ArrayIndexOutOfBoundsException ex) {}
            int n4 = 0;
            for (int l = 0; l < 16; ++l) {
                for (int n5 = 0; n5 < 128; ++n5) {
                    int n6 = array[n5][l];
                    if (n6 >= 0) {
                        RealTimeSequencer.this.getTransmitterList().sendMessage(0xB0 | l | n5 << 8 | n6 << 16, -1L);
                        ++n4;
                    }
                }
                if (array2[l] >= 0) {
                    RealTimeSequencer.this.getTransmitterList().sendMessage(0xC0 | l | array2[l] << 8, -1L);
                }
                if (array2[l] >= 0 || n2 == 0L || n3 == 0L) {
                    RealTimeSequencer.this.getTransmitterList().sendMessage(0xE0 | l | 0x400000, -1L);
                    RealTimeSequencer.this.getTransmitterList().sendMessage(0xB0 | l | 0x4000, -1L);
                }
            }
        }
        
        synchronized void chaseEvents(long n, long n2) {
            byte[][] array = new byte[128][16];
            for (int i = 0; i < this.tracks.length; ++i) {
                if (this.trackDisabled == null || this.trackDisabled.length <= i || !this.trackDisabled[i]) {
                    this.chaseTrackEvents(i, n, n2, true, array);
                }
            }
        }
        
        private long getCurrentTimeMillis() {
            return System.nanoTime() / 1000000L;
        }
        
        private long millis2tick(long n) {
            if (this.divisionType != 0.0f) {
                return (long)((double)n * (double)this.tempoFactor * (double)this.divisionType * (double)this.resolution / 1000.0);
            }
            return MidiUtils.microsec2ticks(n * 1000L, (double)(this.currTempo * this.inverseTempoFactor), this.resolution);
        }
        
        private long tick2millis(long n) {
            if (this.divisionType != 0.0f) {
                return (long)((double)n * 1000.0 / ((double)this.tempoFactor * (double)this.divisionType * (double)this.resolution));
            }
            return MidiUtils.ticks2microsec(n, (double)(this.currTempo * this.inverseTempoFactor), this.resolution) / 1000L;
        }
        
        private void ReindexTrack(int n, long n2) {
            if (n < this.trackReadPos.length && n < this.tracks.length) {
                this.trackReadPos[n] = MidiUtils.tick2index(this.tracks[n], n2);
            }
        }
        
        private boolean dispatchMessage(int n, MidiEvent midiEvent) {
            boolean n2 = false;
            MidiMessage message = midiEvent.getMessage();
            int status = message.getStatus();
            int length = message.getLength();
            if (status == 255 && length >= 2) {
                if (n == 0) {
                    int tempoMPQ = MidiUtils.getTempoMPQ(message);
                    if (tempoMPQ > 0) {
                        if (midiEvent.getTick() != this.ignoreTempoEventAt) {
                            this.setTempoMPQ((float)tempoMPQ);
                            n2 = true;
                        }
                        this.ignoreTempoEventAt = -1L;
                    }
                }
                RealTimeSequencer.this.sendMetaEvents(message);
            }
            else {
                RealTimeSequencer.this.getTransmitterList().sendMessage(message, -1L);
                switch (status & 0xF0) {
                    case 128: {
                        int n3 = ((ShortMessage)message).getData1() & 0x7F;
                        int[] noteOnCache = this.noteOnCache;
                        int n4 = n3;
                        noteOnCache[n4] &= (0xFFFF ^ 1 << (status & 0xF));
                        break;
                    }
                    case 144: {
                        ShortMessage shortMessage = (ShortMessage)message;
                        int n5 = shortMessage.getData1() & 0x7F;
                        if ((shortMessage.getData2() & 0x7F) > 0) {
                            int[] noteOnCache3 = this.noteOnCache;
                            int n7 = n5;
                            noteOnCache3[n7] |= 1 << (status & 0xF);
                            break;
                        }
                        int[] noteOnCache2 = this.noteOnCache;
                        int n6 = n5;
                        noteOnCache2[n6] &= (0xFFFF ^ 1 << (status & 0xF));
                        break;
                    }
                    case 176: {
                        RealTimeSequencer.this.sendControllerEvents(message);
                        break;
                    }
                }
            }
            return n2;
        }
        
        synchronized boolean pump() {
            long n = this.lastTick;
            int n2 = 0;
            long checkPointMillis = this.getCurrentTimeMillis();
            int i;
            int n5;
            do {
                i = 0;
                if (this.needReindex) {
                    if (this.trackReadPos.length < this.tracks.length) {
                        this.trackReadPos = new int[this.tracks.length];
                    }
                    for (int j = 0; j < this.tracks.length; ++j) {
                        this.ReindexTrack(j, n);
                    }
                    this.needReindex = false;
                    this.checkPointMillis = 0L;
                }
                if (this.checkPointMillis == 0L) {
                    checkPointMillis = this.getCurrentTimeMillis();
                    this.checkPointMillis = checkPointMillis;
                    n = this.lastTick;
                    this.checkPointTick = n;
                }
                else {
                    n = this.checkPointTick + this.millis2tick(checkPointMillis - this.checkPointMillis);
                    if (RealTimeSequencer.this.loopEnd != -1L && ((RealTimeSequencer.this.loopCount > 0 && this.currLoopCounter > 0) || RealTimeSequencer.this.loopCount == -1) && this.lastTick <= RealTimeSequencer.this.loopEnd && n >= RealTimeSequencer.this.loopEnd) {
                        n = RealTimeSequencer.this.loopEnd - 1L;
                        n2 = 1;
                    }
                    this.lastTick = n;
                }
                int n3 = 0;
                int k = 0;
                while (k < this.tracks.length) {
                    try {
                        int n4 = this.trackDisabled[k];
                        Track track = this.tracks[k];
                        int index = this.trackReadPos[k];
                        int size = track.size();
                        MidiEvent value;
                        while (i == 0 && index < size && (value = track.get(index)).getTick() <= n) {
                            if (index == size - 1 && MidiUtils.isMetaEndOfTrack(value.getMessage())) {
                                index = size;
                                break;
                            }
                            ++index;
                            if (n4 && (k != 0 || !MidiUtils.isMetaTempo(value.getMessage()))) {
                                continue;
                            }
                            i = this.dispatchMessage(k, value);
                        }
                        if (index >= size) {
                            ++n3;
                        }
                        this.trackReadPos[k] = index;
                    }
                    catch (Exception ex) {
                        if ((ex instanceof ArrayIndexOutOfBoundsException)) {
                            this.needReindex = true;
                            i = 1;
                        }
                    }
                    if (i != 0) {
                        break;
                    }
                    ++k;
                }
                n5 = (n3 == this.tracks.length);
                if (n2 != 0 || (((RealTimeSequencer.this.loopCount > 0 && this.currLoopCounter > 0) || RealTimeSequencer.this.loopCount == -1) && i == 0 && RealTimeSequencer.this.loopEnd == -1L && n5)) {
                    long checkPointMillis2 = this.checkPointMillis;
                    long n6 = RealTimeSequencer.this.loopEnd;
                    if (n6 == -1L) {
                        n6 = this.lastTick;
                    }
                    if (RealTimeSequencer.this.loopCount != -1) {
                        --this.currLoopCounter;
                    }
                    this.setTickPos(RealTimeSequencer.this.loopStart);
                    this.checkPointMillis = checkPointMillis2 + this.tick2millis(n6 - this.checkPointTick);
                    this.checkPointTick = RealTimeSequencer.this.loopStart;
                    this.needReindex = false;
                    i = 0;
                    n2 = 0;
                    n5 = false;
                }
                else {
                    continue;
                }
            } while (i != 0);
            return n5;
        }
    }
}
